/*
 *  buddy.c
 *
 *  Some functions to cope with buddy IM and buddy list.
 *
 *  Copyright (C) 2008  Huang Guan
 *
 *  2008-01-31 Created.
 *
 *  Description: This file mainly includes the functions about 
 *               Buddy IM 
 *               Buddy List
 *               Buddy Information
 *               Buddy Status
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#ifdef __WIN32__
#include <winsock.h>
#include <wininet.h>
#else
#include <sys/socket.h>
#include <arpa/inet.h>
#endif

#include "myqq.h"
#include "packet.h"
#include "cs.h"
#include "debug.h"
#include "util.h"



//
//		** BUDDY MANAGEMENT **
//
//return my buddy information pointer
//if not found, create a new one
myqq_buddy_t* buddy_get2( int qid, uint uid, int create )
{
	int i;
	if( uid == 0 ) return NULL;
	//we don't know how many threads we've created, they may run here together and cause any race
	//conditions
	pthread_mutex_lock( &MYQQ(qid).buddy_mutex );
	for( i=0; i<MYQQ(qid).buddy_count; i++ )
	{
		if( MYQQ(qid).buddy_list[i]->uid == uid ){
			pthread_mutex_unlock( &MYQQ(qid).buddy_mutex );
			return MYQQ(qid).buddy_list[i];
		}
	}
	//not found in buddy list
	if( !create ){
		pthread_mutex_unlock( &MYQQ(qid).buddy_mutex );
		return NULL;
	}
	//create it
	if( MYQQ(qid).buddy_count >= MAX_BUDDY )
	{
		DBG("IMPOSSIBLE!");
		exit(1);
	}
	i = MYQQ(qid).buddy_count;
	MYQQ(qid).buddy_list[i] = (myqq_buddy_t*)malloc( sizeof(myqq_buddy_t) );
	memset( MYQQ(qid).buddy_list[i], 0, sizeof(myqq_buddy_t) );
	sprintf( MYQQ(qid).buddy_list[i]->nick, "%d", uid );
	MYQQ(qid).buddy_list[i]->uid = uid;
	MYQQ(qid).buddy_count ++;
	pthread_mutex_unlock( &MYQQ(qid).buddy_mutex );
	return MYQQ(qid).buddy_list[i];
}

//old style function
myqq_buddy_t* buddy_get( int qid, uint uid )
{
	return buddy_get2( qid, uid, 1 );
}

//
void buddy_remove( int qid, uint uid )
{
	int i;
	if( uid == 0 ) return;
	pthread_mutex_lock( &MYQQ(qid).buddy_mutex );
	for( i=0; i<MYQQ(qid).buddy_count; i++ )
	{
		if( MYQQ(qid).buddy_list[i]->uid == uid ){
			//remove it
			myqq_buddy_t* tmp = MYQQ(qid).buddy_list[i];
			if( i != MYQQ(qid).buddy_count-1 )
				MYQQ(qid).buddy_list[i] = MYQQ(qid).buddy_list[MYQQ(qid).buddy_count-1];
			MYQQ(qid).buddy_count--;
			free( tmp );
			pthread_mutex_unlock( &MYQQ(qid).buddy_mutex );
			return;
		}
	}
	DBG("uid: %d not found.", uid );
	pthread_mutex_unlock( &MYQQ(qid).buddy_mutex );
}

static void buddy_set_all_offline( int qid )
{
	int i;
	pthread_mutex_lock( &MYQQ(qid).buddy_mutex );
	for( i=0; i<MYQQ(qid).buddy_count; i++ )
	{
		MYQQ(qid).buddy_list[i]->mode = LM_OFFLINE;
	}
	pthread_mutex_unlock( &MYQQ(qid).buddy_mutex );
}

//****************************************************************************************
//
//				Buddy IM
//
//
typedef struct __attribute__((packed)) myqq_prot_recv_im_common_s {
	ushort		sender_ver;
	uint		sender_uid;
	uint		recver_uid;
	char		session_key_md5[16];
	ushort		text_type;
}myqq_prot_recv_im_common_t;

typedef struct __attribute__((packed)) myqq_prot_recv_im_text_head_s {
	ushort		msg_seqno;
	time_t		send_time;
	ushort		sender_icon;
	char		unknown1[3];	//always 0
	char		msg_attr;	//normally 1
	char		unknown2[4];	//always 0
	char		msg_type;
}myqq_prot_recv_im_text_head_t;

typedef struct __attribute__((packed)) myqq_prot_recv_im_text_tail_s {
	char		font_size;
	char		font_color[3];//RGB
	char		unknown;	//always 0
	char		char_type;	//0x86:GB2312  0x00:ASCII
	char		unknown2;
	char		font_name[1];
}myqq_prot_recv_im_text_tail_t;


typedef struct __attribute__((packed)) myqq_prot_recv_im_text_s {
	myqq_prot_recv_im_common_t*		im_common;
	myqq_prot_recv_im_text_head_t*	im_text_head;
	char*					im_text;	//length we don't known. it ends with 0x00
	myqq_prot_recv_im_text_tail_t*	im_text_tail;
}myqq_prot_recv_im_text_t;

//Delegater
EXPORT void myqq_delegate_buddy_msg( int qid, messenger msgr )
{
	MYQQ(qid).buddy_msgr = msgr;
}

static void buddy_recv_text_im( int qid, char* data, int len)
{
	myqq_prot_recv_im_text_t text;
	int i = 0;
	text.im_common = (myqq_prot_recv_im_common_t*) ( data+i );
	i += sizeof(myqq_prot_recv_im_common_t);
	text.im_text_head = (myqq_prot_recv_im_text_head_t*)(data+i);
	i += sizeof(myqq_prot_recv_im_text_head_t);
	text.im_text = (char*)(data+i);
	time_t t = ntohl( text.im_text_head->send_time );
	while( i<len && *(data+i)!=0 ) i++;
	i += 1;
	text.im_text_tail = (myqq_prot_recv_im_text_tail_t*)(data+i);
	//Print the message we recved.
	if( MYQQ(qid).buddy_msgr )
	{
		uint uid = ntohl( text.im_common->sender_uid );
		myqq_buddy_t * b = buddy_get2( qid, uid, 0 );
		if( b == NULL ) //don't know??
		{	//maybe a stranger talks to us.
			b = buddy_get2( qid, uid, 1 );
			if( b == NULL ) return; //seems impossible!! right?
			buddy_get_info( qid, uid ); //see who is he
		}
		MYQQ(qid).buddy_msgr( uid, b->nick, t, text.im_text );
	}else{
		MSG("Message from [%d], len: %d, date: %s %s\n", ntohl( text.im_common->sender_uid ),
			(int)text.im_text_tail-(int)text.im_text-1, ctime( &t ), text.im_text );
	}
}

//cope with buddy's message
void buddy_recv_im( int qid, char* data, int len )
{
	myqq_prot_recv_im_common_t* common = (myqq_prot_recv_im_common_t*)data;
	switch( ntohs(common->text_type) ){
	case QQ_NORMAL_IM_TEXT:	// 0x000b
		buddy_recv_text_im( qid, data, len );
		break;
	case QQ_NORMAL_IM_FILE_REQUEST_TCP:	// 0x0001
		DBG("QQ_NORMAL_IM_FILE_REQUEST_TCP");
		break;
	case QQ_NORMAL_IM_FILE_APPROVE_TCP:	// 0x0003
		DBG("QQ_NORMAL_IM_FILE_APPROVE_TCP");
		break;
	case QQ_NORMAL_IM_FILE_REJECT_TCP:	// 0x0005
		DBG("QQ_NORMAL_IM_FILE_REJECT_TCP");
		break;
	case QQ_NORMAL_IM_FILE_REQUEST_UDP:	// 0x0035
		DBG("QQ_NORMAL_IM_FILE_REQUEST_UDP");
		break;
	case QQ_NORMAL_IM_FILE_APPROVE_UDP:	// 0x0037
		DBG("QQ_NORMAL_IM_FILE_APPROVE_UDP");
		break;
	case QQ_NORMAL_IM_FILE_REJECT_UDP:	// 0x0039
		DBG("QQ_NORMAL_IM_FILE_REJECT_UDP");
		break;
	case QQ_NORMAL_IM_FILE_NOTIFY:	// 0x003b
		DBG("QQ_NORMAL_IM_FILE_NOTIFY");
		break;
	case QQ_NORMAL_IM_FILE_PASV:	// 0x003f
		DBG("QQ_NORMAL_IM_FILE_PASV");
		break;
	case QQ_NORMAL_IM_FILE_CANCEL:	// 0x0049
		DBG("QQ_NORMAL_IM_FILE_CANCEL");
		break;
	case QQ_NORMAL_IM_FILE_EX_REQUEST_UDP:	// 0x81
		DBG("QQ_NORMAL_IM_FILE_EX_REQUEST_UDP");
		break;
	case QQ_NORMAL_IM_FILE_EX_REQUEST_ACCEPT:	// 0x83
		DBG("QQ_NORMAL_IM_FILE_EX_REQUEST_ACCEPT");
		break;
	case QQ_NORMAL_IM_FILE_EX_REQUEST_CANCEL:	// 0x85
		DBG("QQ_NORMAL_IM_FILE_EX_REQUEST_CANCEL");
		break;
	case QQ_NORMAL_IM_FILE_EX_NOTIFY_IP:	// 0x87
		DBG("QQ_NORMAL_IM_FILE_EX_NOTIFY_IP");
		break;
	default:
		DBG("Unknown buddy im type: 0x%x", ntohs(common->text_type) );
	}
}

//****************************************************************************************
//
//				Buddy Information
//
//
typedef struct __attribute__((packed)) myqq_buddy_status_s {
	uint		uid;
	char		unknown1;	//always 0
	uint		ip;
	ushort		port;
	ushort		status;	//0x0A: Normal  0x14: OffLine  0x1E: Away
	ushort		version;
	char		session_key_md5[16];
	char		unknown[2];	//0
	uint		my_uid;
}myqq_buddy_status_t;

//Get buddy information
void buddy_get_info( int qid, uint uid )
{
	char str_uid[16];
	int len;
	len = sprintf( str_uid, "%d", uid );
	//Send Command to Server, need acknowledge packet
	cs_send_cmd( qid, QQ_CMD_GET_USER_INFO, str_uid, len, 1 );
}


//process buddy information
void buddy_recv_info( int qid, char* data, int len )
{
	int i, j, k, uid, v;
#define GET_NEXT do{i++;}while( i<len && data[i]!=0x1E ); data[i]=0; 
	//hex_dump( data, len );
	//get uid first
	i = j = v = 0;
	GET_NEXT
	uid = atoi( &data[j] ); 
	myqq_buddy_t* buddy = buddy_get2( qid, uid, 1 );
	// we dont interest in this infor.
	strcpy( buddy->info[v++], &data[j] );	//save uid info.
	j = i + 1;
	while( i<len )
	{
		GET_NEXT
		k = i - j;
		strcpy( buddy->info[v++], &data[j] );
		j = i + 1;
		if( v >= MAX_BUDDY_INFO ) break;
	}
}


//Buddy Status Delegate
//2008-2-12 15:46:34 Huang Guan
//typedef void (*newstatus) ( uint uid, char* nick, int status );
EXPORT void myqq_delegate_buddy_status( int qid, newstatus msgr )
{
	MYQQ(qid).status_msgr = msgr;
}

//buddy status changed
void buddy_recv_status( int qid, char * data, int len )
{
	myqq_buddy_status_t* st = (myqq_buddy_status_t*)data;
	//0x0A: Normal  0x14: OffLine  0x1E: Away
	int uid = ntohl( st->uid );
	myqq_buddy_t *b = buddy_get2( qid, uid, 0 );
	if( b == NULL ) return;
	//if( ntohl( st->my_uid ) != MYQQ(qid).uid )
	//	DBG("uid not equal.");
	switch( ntohs( st->status ) ){
	case 0x0A:
		//MSG("UID:%d is Onlilne.\n", uid );
		b->mode = LM_NORMAL;
		break;
	case 0x14:
		//MSG("UID:%d is Offline.\n", uid );
		b->mode = LM_OFFLINE;
		break;
	case 0x1E:
		//MSG("UID:%d is away.\n", uid );
		b->mode = LM_AWAY;
		break;
	}
	if( MYQQ(qid).status_msgr )
		MYQQ(qid).status_msgr( uid, b->nick, b->mode );
}


//send a text message to buddy
void buddy_send_text_im( int qid, uint uid, char* msg )
{
	char *data = (char*) malloc( MAX_PACKET_SIZE );
	static const char tail[]={ 
		0x0A,		0x00,0x00,0x00,	0x00,	//font size, font rgb, zero
		0x86,0x02,					//character type
		0xCB,0xCE,0xCC,0xE5,0x0D };			//font name
	int msg_len = strlen( msg );
	if( msg_len > 65536/2 )
	{
		MSG("Message too long.\n");
		return;
	}
	myqq_prot_recv_im_common_t common;
	myqq_prot_recv_im_text_head_t head;
	memset( &head, 0, sizeof( head ) );
	common.sender_ver = htons( MYQQ(qid).version );
	common.sender_uid = htonl( MYQQ(qid).uid );
	common.recver_uid = htonl( uid );
	memcpy( common.session_key_md5, MYQQ(qid).session_md5, 16 );
	common.text_type = htons( QQ_NORMAL_IM_TEXT );
	head.msg_seqno = htons( MYQQ(qid).seqno );
	head.send_time = htonl( time(NULL) );
	head.sender_icon = htons( 0 );
	head.msg_attr = 1;
	head.msg_type = QQ_IM_TEXT; //QQ_IM_AUTO_REPLY
	int cb = 0;
	memcpy( data + cb, &common, sizeof( myqq_prot_recv_im_common_t ) );
	cb += sizeof( myqq_prot_recv_im_common_t );
	memcpy( data + cb, &head, sizeof( myqq_prot_recv_im_text_head_t ) );
	cb += sizeof( myqq_prot_recv_im_text_head_t );
	memcpy( data + cb, msg, msg_len + 1 );	//put 0x00 at the end
	cb += msg_len + 1;
	memcpy( data + cb, tail, sizeof( tail ) );
	cb += sizeof(tail);
	im_send( qid, uid, data, cb );
	free( data );
}




//****************************************************************************************
//
//				Buddy List
//
//

typedef struct __attribute__((packed)) myqq_buddy_list_item_s {
	uint		uid;
	ushort		face;		//face index
	char		age;
	char		sex;
	char		nick_len;	//
	char		nick[1];
}myqq_buddy_list_item_t;

typedef struct __attribute__((packed)) myqq_buddy_online_item_s {
	uint		uid;
	char		unknown;
	uint		ip;
	ushort		port;
	char		unknown2;
	char		mode;
	char		unknown3[2];
	char		session_key[16];	//for P2P chating.
	char		unknown4[2];
	char		qqshow;
	char		flag;
	char		unknown5[2];
	char		separator;		//0x00
}myqq_buddy_online_item_t;

EXPORT void myqq_update_buddy_list( int qid )
{
	buddy_get_list( qid, 0 );
	buddy_get_online_list( qid, 0 );
}

//get buddy list
void buddy_get_list( int qid, int pos )
{
	char data[3];
	memset( data, 0, 3 ); //set zero obviously
	ushort tmp = pos;
	tmp = htons( tmp );
	memcpy( data, &tmp, 2 );
	data[2] = 0; //not sorted.
	cs_send_cmd( qid, QQ_CMD_GET_FRIENDS_LIST, data, 3, 1 );
}

#ifndef MIN
#define MIN( a, b ) ( a > b ? b : a )
#endif
void buddy_recv_list( int qid, char* data, int len )
{
	ushort pos;
	myqq_buddy_list_item_t *item;
	memcpy( &pos, data, 2 );
	pos = ntohs( pos );
	int i;
	char* cur;
	for( i = 2; i < len; )
	{
		cur = data + i;
		item = (myqq_buddy_list_item_t*) cur;
		myqq_buddy_t* b = buddy_get( qid, ntohl( item->uid ) );
		b->sex = item->sex;
		b->age = item->age;
		b->face = ntohs( item->face );
		//cut the name, or it would be like this: 'Flash2342'
		sprintf( b->nick, "%.*s", MIN( item->nick_len, MAX_NAME_LEN ), item->nick );
		//get qq show flag and mobile flag
		i += sizeof( myqq_buddy_list_item_t ) - 1; //nick[1]
		i += item->nick_len;
		i += 2; //skip two unknown bytes;
		cur = data + i;
		b->qqshow = *cur;
		cur ++;
		b->flag = *cur;
		i += 2;
	}
	if( pos != 0xffff )
		buddy_get_list( qid, pos );
}


void buddy_get_online_list( int qid, int pos )
{
	char data[5];
	memset( data, 0, 5 ); //set zero obviously
	data[0] = 0x02;  //0x02 or 0x03
	if( pos < 0 || pos > 255 ) DBG("Impossible: pos=%d", pos );
	data[1] = pos;
	cs_send_cmd( qid, QQ_CMD_GET_FRIENDS_ONLINE, data, 3, 1 );
}


void buddy_recv_online_list( int qid, char* data, int len )
{
	myqq_buddy_online_item_t *item;
	int pos;
	pos = data[0];
	//set all buddy offline.
	buddy_set_all_offline( qid );
	for( item = (myqq_buddy_online_item_t*)&data[1];
		item < (myqq_buddy_online_item_t*)&data[len]; item ++ )
	{
		myqq_buddy_t* b = buddy_get( qid, ntohl( item->uid ) );
		DBG("UID: %d Nick: %s is online.", b->uid, b->nick );
		b->ip = ntohl( item->ip );
		b->port = ntohs( item->port );
		if( item->mode == 0x0A )
			b->mode = LM_NORMAL;
		else if( item->mode == 0x1E )
			b->mode = LM_AWAY;
		memcpy( b->session_key, item->session_key, 16 );
		b->qqshow = item->qqshow;
		b->flag = item->flag;
	}
}

//
/*   The output buf looks like this:
L8bit  L16bit        L16bit        L32bit
1      357339036     online		Xiao xia
2      273310179     offline	Huang Guan
*/
//Note: size must be big enough
EXPORT int myqq_get_buddy_list( int qid, char* buf, int size, char online )
{
	int i, len = 0;
	char mode_str[4][8] = {"OffLine", "OnLine", "Hidden", "Away" };
	//avoid editing the array
	buf[0] = 0;
	pthread_mutex_lock( &MYQQ(qid).buddy_mutex );
	int ln = 1;
	for( i=0; i<MYQQ(qid).buddy_count; i++ )
	{
		myqq_buddy_t * b = MYQQ(qid).buddy_list[i];
		if( online && b->mode == LM_OFFLINE ) continue;
		len = sprintf( buf, "%s%-8d%-16d%-16s%-16.64s\n", buf, ln ++, b->uid, 
			mode_str[(int)b->mode], util_escape( b->nick ) );
		if( len + 80 > size ) break;
	}
	pthread_mutex_unlock( &MYQQ(qid).buddy_mutex );
	return len;
}

//
EXPORT int myqq_send_im_to_buddy( int qid, uint uid, char* msg, int wait )
{
	buddy_send_text_im( qid, uid, msg );
	if( wait )
	{
		//wait for 10 seconds to see if all the packets have been sent.
		if( myqq_wait( qid, 15 ) < 0 )
			return -1;
	}
	return 0;
}

//
EXPORT const char* myqq_get_buddy_name( int qid, uint uid )
{
	myqq_buddy_t* b = buddy_get2( qid, uid, 0 );
	if( b == NULL )
		return "None";
	else
		return b->nick;
}


//interface for getting buddy information
/* Output style: 
*/
EXPORT int myqq_get_buddy_info( int qid, uint uid, char* buf, int size )
{
	myqq_buddy_t *b = buddy_get2( qid, uid, 0 );
	if( size < 256 )
		return -EFAULT;
	if( b == NULL )
		return -EFAULT;
	int len, ip = htonl(b->ip);
	len = sprintf( buf,  "BuddyName           %-32s\n"
				"UserID              %-32d\n"
				"IPAddress           %-32s\n"
				"Port                %-32d\n"
				"Face                %-32d\n"
				"Age                 %-32d\n"
				"Sex                 %-32s\n"
				"Mode                %-32s\n",
		b->nick, b->uid, inet_ntoa( *(struct in_addr*)&ip ), b->port, b->face, b->age,
		(b->sex==0x00)?"Male": (b->sex==0x01)?"Female":"Asexual",
		(b->mode==LM_OFFLINE)?"OffLine": (b->mode==LM_AWAY)?"Away" : "Online" );
	return len;
}



