/*
 *  group.c
 *
 *  Some functions to cope with group IM and group information.
 *
 *  Copyright (C) 2008  Huang Guan
 *
 *  2008-01-31 Created.
 *
 *  Description: This file mainly includes the functions about 
 *               Group IM 
 *               Group List
 *               Group Information
 *               Group Member List
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#ifdef __WIN32__
#include <winsock.h>
#else
#include <sys/socket.h>
#include <arpa/inet.h>
#endif

#include "myqq.h"
#include "packet.h"
#include "protocol.h"
#include "cs.h"
#include "debug.h"
#include "util.h"


//
//		** GROUP MANAGEMENT **
//
//return my group information pointer
//if not found, create a new one
myqq_group_t* group_get2( int qid, uint uid, int create )
{
	int i;
	if( uid == 0 ) return NULL;
	//we don't know how many threads we've created, they may run here together and cause any race
	//conditions
	pthread_mutex_lock( &MYQQ(qid).group_mutex );
	for( i=0; i<MYQQ(qid).group_count; i++ )
	{
		if( MYQQ(qid).group_list[i]->int_uid == uid ){
			pthread_mutex_unlock( &MYQQ(qid).group_mutex );
			return MYQQ(qid).group_list[i];
		}
	}
	//not found in group list
	if( !create ){
		pthread_mutex_unlock( &MYQQ(qid).group_mutex );
		return NULL;
	}
	//create it
	if( MYQQ(qid).group_count >= MAX_BUDDY )
	{
		DBG("IMPOSSIBLE!");
		exit(1);
	}
	i = MYQQ(qid).group_count;
	MYQQ(qid).group_list[i] = (myqq_group_t*)malloc( sizeof(myqq_group_t) );
	memset( MYQQ(qid).group_list[i], 0, sizeof(myqq_group_t) );
	pthread_mutex_init( &MYQQ(qid).group_list[i]->member_mutex, NULL );
	MYQQ(qid).group_list[i]->int_uid = uid;
	sprintf( MYQQ(qid).group_list[i]->name, "%d", uid );
	MYQQ(qid).group_count ++;
	pthread_mutex_unlock( &MYQQ(qid).group_mutex );
	return MYQQ(qid).group_list[i];
}

myqq_group_t* group_get( int qid, uint uid )
{
	return group_get2( qid, uid, 1 );
}

void group_remove( int qid, uint uid )
{
	int i;
	if( uid == 0 ) return;
	pthread_mutex_lock( &MYQQ(qid).group_mutex );
	for( i=0; i<MYQQ(qid).group_count; i++ )
	{
		if( MYQQ(qid).group_list[i]->ext_uid == uid ){
			//remove it
			myqq_group_t* tmp = MYQQ(qid).group_list[i];
			if( i != MYQQ(qid).group_count-1 )
				MYQQ(qid).group_list[i] = MYQQ(qid).group_list[MYQQ(qid).group_count-1];
			MYQQ(qid).group_count--;
			free( tmp );
			pthread_mutex_unlock( &MYQQ(qid).group_mutex );
			return;
		}
	}
	DBG("uid: %d not found.", uid );
	pthread_mutex_unlock( &MYQQ(qid).group_mutex );
}

myqq_group_member_t* group_get_member( int qid, myqq_group_t* g, uint uid )
{
	int i;
	if( g == NULL || uid == 0 ) return NULL;
	pthread_mutex_lock( &g->member_mutex );
	for( i=0; i<g->member_count; i++ )
	{
		if( g->member_list[i].uid == uid ){
			pthread_mutex_unlock( &g->member_mutex );
			return &g->member_list[i];
		}
	}
	//not found in group member list
	//create it
	if( g->member_count >= MAX_GROUP_MEMBER )
	{
		DBG("IMPOSSIBLE!"); exit(1);
	}
	i = g->member_count;
	g->member_list[i].uid = uid;
	sprintf( g->member_list[i].nick, "%d", uid );
	g->member_count ++;
	pthread_mutex_unlock( &g->member_mutex );
	return &g->member_list[i];
}


//****************************************************************************************
//
//				GROUP IM
//
//
typedef struct __attribute__((packed)) myqq_prot_recv_im_group_common_s {
	uint		ext_uid;	//int_uidΪڲȺ  ext_uidΪⲿȺ
	char		group_type;	//0x01 ̶Ⱥ  0x02 ʱȺ
	uint		uid;	//qq
	char		data[2];
}myqq_prot_recv_im_group_common_t;

//ȺϢͷ
typedef struct __attribute__((packed)) myqq_prot_recv_im_group_head_s {
	ushort		msg_seqno;
	time_t		send_time;
	char		unknown[4];	//always 0
	ushort		msg_len;	//Ϣ
}myqq_prot_recv_im_group_head_t;

typedef struct __attribute__((packed)) myqq_prot_recv_im_group_tail_s {
	char		font_size;
	ushort		font_color;
	char		unknown;	//always 0
	char		char_type;	//0x86:GB2312  0x00:ASCII
	char		unknown2;
	char		font_name[1];
}myqq_prot_recv_im_group_tail_t;


typedef struct __attribute__((packed)) myqq_prot_recv_im_group_s {
	myqq_prot_recv_im_t*			im_head;
	myqq_prot_recv_im_group_common_t*	im_common;
	myqq_prot_recv_im_group_head_t*	im_text_head;
	char*					im_text;	//length we don't known. it ends with 0x00
	myqq_prot_recv_im_group_tail_t*	im_text_tail;
}myqq_prot_recv_im_group_t;


//Delegater
EXPORT void myqq_delegate_group_msg( int qid, group_messenger msgr )
{
	MYQQ(qid).group_msgr = msgr;
}


void group_recv_im( int qid, char* data, int len, int cmd )
{
	myqq_prot_recv_im_group_t text;
	int i = 0;
	text.im_head = (myqq_prot_recv_im_t*) data;
	i += sizeof(myqq_prot_recv_im_t);
	text.im_common = (myqq_prot_recv_im_group_common_t*) ( data+i );
	i += sizeof(myqq_prot_recv_im_group_common_t);
	text.im_text_head = (myqq_prot_recv_im_group_head_t*)(data+i);
	i += sizeof(myqq_prot_recv_im_group_head_t);
	if( ntohs(text.im_head->type) != QQ_RECV_IM_UNKNOWN_QUN_IM )
		i += 10;
	text.im_text = (char*)(data+i);
	time_t t = ntohl( text.im_text_head->send_time );
	while( i<len && text.im_text[i]!=0 ) i++;
	i += 1;
	text.im_text_tail = (myqq_prot_recv_im_group_tail_t*)(data+i);
	if( MYQQ(qid).group_msgr )
	{
		uint uid, int_uid;
		int_uid = ntohl( text.im_head->int_uid );
		uid = ntohl( text.im_common->uid );
		myqq_group_t *g = group_get( qid, int_uid );	
		myqq_group_member_t* m = group_get_member( qid, g, uid );
		MYQQ(qid).group_msgr( uid, m->nick, int_uid, g->name, t, text.im_text );
	}else{
		//Print the message we recved.
		MSG("GROUP[%d] Message UID[%d], len: %d, date: %s %s\n",
			ntohl( text.im_common->ext_uid ), ntohl( text.im_common->uid ),
			(int)text.im_text_tail-(int)text.im_text-1, ctime( &t ), text.im_text );
	}
}

//group im: send message to group
void group_send_im( int qid, uint int_uid, char* msg )
{
	static const char tail[]={ 
		0x0A,		0x00,0x00,0x00,	0x00,	//font size, font rgb, zero
		0x86,0x02,					//character type
		0xCB,0xCE,0xCC,0xE5,0x0D };			//font name
	int len = strlen( msg );
	char* data = malloc( len + 32 );
	data[0] = 0x0A;	//Group send IM sub-cmd.
	int tmp = htonl( int_uid );
	memcpy( &data[1], &tmp, 4 );
	tmp = htons( len );
	memcpy( &data[5], &tmp, 2 );
	memcpy( &data[7], msg, len ); //ends with 0x00
	len += 7;
	memcpy( &data[len], tail, sizeof(tail) );
	len += sizeof(tail);
	cs_send_cmd( qid, QQ_CMD_GROUP_CMD, data, len, 1 );
	free( data );
}

//send message to group
EXPORT int myqq_send_im_to_group( int qid, uint int_uid, char* msg, int wait )
{
	group_send_im( qid, int_uid, msg );
	if( wait )
	{
		if( myqq_wait( qid, 15 ) < 0 )
			return -1;
	}
	return 0;
}

//****************************************************************************************
//
//				GROUP LIST AND GROUP INFORMATION
//
//
typedef struct __attribute__((packed)) myqq_prot_group_item_s {
	uint		uid;
	char		type;	//0x01: buddy  0x04: group
	char		unknown;	//sometime 0x00, sometime 0x01
}myqq_prot_group_item_t;

//get group list and also buddy list
//we are not interested in buddy list.
//becase we also get it with another better command.
void group_get_list( int qid, uint pos )
{
	char data[10];
	memset( data, 0, 10 ); //set zero obviously
	data[0] = 0x01;  	//0x01: download  0x02: upload
	data[1] = 0x02;	//unknown
	pos = htonl( pos );
	memcpy( data+6, &pos, 4 );
	cs_send_cmd( qid, QQ_CMD_GET_ALL_LIST_WITH_GROUP, data, 10, 1 );
}

//If you want to see what the data like, just add hex_dump( data, len );
void group_recv_list( int qid, char* data, int len )
{
	if( data[0] != 0x01 || data[1] != 0x00 )
	{
		DBG("Wrong data.");	return;
	}
	myqq_prot_group_item_t* item; 
	int pos;
	pos = ntohl( *(uint*)&data[6] );
	for( item = (myqq_prot_group_item_t*)&data[10];
		item < (myqq_prot_group_item_t*)&data[len]; item ++ )
	{
		if( item->type == 0x04 )
		{
			//the group id is internal id
			myqq_group_t* b = group_get( qid, ntohl( item->uid ) );
			if( b->ext_uid == 0 ){
				group_get_info( qid, b->int_uid );
			}
		}
	}
	if( pos )
		group_get_list( qid, pos );
}


//
/*   The output buf looks like this:
L8bit  L16bit        L16bit        L32bit
1      467788923     12118724	Xinxing Experimental School
2      489234223     13223423	SGOS2007
*/
//Note: size must be big enough
EXPORT int myqq_get_group_list( int qid, char* buf, int size )
{
	int i, len = 0, ln=1;
	//avoid editing the array
	buf[0] = 0;
	pthread_mutex_lock( &MYQQ(qid).group_mutex );
	for( i=0; i<MYQQ(qid).group_count; i++ )
	{
		myqq_group_t * g = MYQQ(qid).group_list[i];
		len = sprintf( buf, "%s%-8d%-16d%-16d%-16.64s\n", buf, ln ++, g->int_uid, 
			g->ext_uid, util_escape( g->name ) );
		if( len + 80 > size ) break;
	}
	pthread_mutex_unlock( &MYQQ(qid).group_mutex );
	return len;
}

/*   The output buf looks like this:
L8bit  L16bit        L16bit        L32bit
1      357339036     Manager	Xiaoxia
2      273310179     Fellow		Huang Guan
*/
//Note: size must be big enough
EXPORT int myqq_get_group_member_list( int qid, uint int_uid, char* buf, int size, char online )
{
	myqq_group_t * g = group_get( qid, int_uid );
	//Hope the group won't get removed while we are using it. 
	int i, len, ln = 1;
	buf[0] = 0;
	for( i=0; i<g->member_count; i++ )
	{
		myqq_group_member_t * m = &g->member_list[i];
		if( online && m->status == ST_OFFLINE ) continue;
		len = sprintf( buf, "%s%-8d%-16d%-16s%-16.64s\n", buf, ln++, m->uid, 
			(m->role&1)?"Admin":"Fellow", util_escape( m->nick ) );
		if( len + 80 > size )
			break;
	}
	return len;
}


void group_get_online_list( int qid, uint int_uid )
{
	char data[5];
	data[0] = 0x0B;	//group sub-cmd: get online group member list
	int tmp = htonl( int_uid );
	memcpy( &data[1], &tmp, 4 );
	cs_send_cmd( qid, QQ_CMD_GROUP_CMD, data, 5, 1 );
}

void group_get_members_info( int qid, uint int_uid )
{
	myqq_group_t *g = group_get( qid, int_uid );
	int len;
	char* data = (char*)malloc( MAX_PACKET_SIZE );
	int i, *cur, *uid_start, num_all = g->member_count, num_send;
	myqq_group_member_t * m = g->member_list;
	data[0] = 0x0C;	//group sub-cmd: get online group member list
	cur = (int*)&data[1];
	*(cur++) = htonl( int_uid );
	//add members' uids
	uid_start = cur;
	while( num_all > 0 )
	{
		if( num_all > 50 )
			num_send = 50;
		else
			num_send = num_all;
		len = 5 + num_send * sizeof(uint);
		i = num_send;
		while( i -- > 0 ){
			* (cur++) = htonl( m->uid );
			m ++;
		}
		cs_send_cmd( qid, QQ_CMD_GROUP_CMD, data, len, 1 );
		cur = uid_start;
		num_all -= num_send;
	}
}

//update group list & infor
EXPORT void myqq_update_group_list( int qid )
{
	group_get_list( qid, 0 );
}

//get group information
void group_get_info( int qid, uint int_uid )
{
	char data[5];
	data[0] = 0x04;	//group sub-cmd: get group information
	*(int*)&data[1] = htonl( int_uid );
	cs_send_cmd( qid, QQ_CMD_GROUP_CMD, data, 5, 1 );
}

//
typedef struct __attribute__((packed)) myqq_prot_group_info_s {
	uint		int_uid;
	uint		ext_uid;
	char		type;		//0x01 normal  0x02 temporary
	char		unknown1[4];	//
	uint		mgr_uid;	//group mananger user;
	char		auth_type;	//
	char		unknown[6];
	uint		category;
	ushort		max_members;
	char		unknown2[5];
}myqq_prot_group_info_t;
//recved group member info
typedef struct __attribute__((packed)) myqq_prot_group_member_s {
	uint		uid;
	char		org;
	char		role;
}myqq_prot_group_member_t;

void group_recv_info( int qid, char* data, int len )
{
	char cmd = data[0];
	int ret, int_uid;
	ret = data[1];
	int_uid = ntohl( *(uint*)&data[2] );
	if( ret != 0x0 )
	{
		DBG("Group command returned failure. Internal ID: %d", int_uid );
		return;
	}
	switch( cmd ){
	case 0x04:	//get group info
	{
		int slen;
		char* cur;
		myqq_prot_group_info_t* info = (myqq_prot_group_info_t*)&data[2];
		myqq_group_t *g = group_get( qid, int_uid );
		if( g == NULL ) return;
		g->ext_uid = ntohl( info->ext_uid );
		g->type = info->type;
		g->mgr_uid = ntohl( info->mgr_uid );
		g->auth_type = info->auth_type;
		g->category = ntohl( info->category );
		g->max_members = ntohs( info->max_members );
		cur = data + 2 + sizeof( myqq_prot_group_info_t );
		//group name
		slen = *cur ++;
		//will remove the prevous name.
		sprintf( g->name, "%.*s", len, cur );
		cur += slen;
		cur += 2;	//skip two unkown bytes;
		slen = *cur ++;
		strncpy( g->notice, cur, slen );
		cur += slen;
		slen = *cur ++;
		strncpy( g->desc, cur, slen );
		cur += slen;
		//group member, separated by 0x00
		myqq_prot_group_member_t *m;
		g->member_count = 0;
		for( m = ( myqq_prot_group_member_t* )cur;
			m < ( myqq_prot_group_member_t* ) &data[len]; m ++ )
		{
			myqq_group_member_t *u = group_get_member( qid, g, ntohl( m->uid ) );
			if( u == NULL ){
				DBG("IMPOSSIBLE."); continue;
			}
			u->role = m->role;
			u->org = m->org;
			sprintf( u->nick, "%d", u->uid );
		}
		//refresh the online members
		group_get_online_list( qid, int_uid );
		break;
	}
	case 0x0A:	//send group message
		break;
	case 0x0B:	//online members
	{
		uint *uid = (uint*)&data[7];	//skip one byte
		myqq_group_t *g = group_get2( qid, int_uid, 0 );
		if( g )
		{
			int i;
			//clear first
			for( i=0; i<g->member_count; i++ )
				g->member_list[i].status = ST_OFFLINE;
			for( ; uid < (uint*)&data[len]; uid ++ )
			{
				//find the member in group and set mode to be online.
				myqq_group_member_t *u = group_get_member( qid, g, ntohl( *uid ) );
				u->status = ST_ONLINE;
			}
			//is it time to refresh member info?
			group_get_members_info( qid, int_uid );
		}
		break;
	}
	case 0x0C:	//group member information
	{
		//a new kind of getting data without using structure.
		char* cur;
		uint uid;
		int str_len;
		myqq_group_t *g = group_get2( qid, int_uid, 0 );
		if( g == NULL )
		{
			DBG("IMPOSSIBLE. g==NULL"); exit(1);
		}
		for( cur = &data[6]; cur < &data[len]; )
		{
			uid = ntohl( *(uint*) cur );
			cur += 4;
			myqq_group_member_t* u = group_get_member( qid, g, uid );
			u->face = ntohs( *(ushort*)cur );
			cur += 2;
			u->age = * (cur++);
			u->sex = * (cur++);
			str_len = * (cur++);
			//will remove the prevous name.
			sprintf( u->nick, "%.*s", str_len, cur );
			//u->nick[str_len] = 0;
			cur += str_len;
			cur += 2;	//skip two unknown bytes
			u->qqshow = * (cur++);
			u->flag = * (cur++);
		}
		break;
	}
	default:
		DBG("group cmd %d, not implemented.", cmd );
	}
}


EXPORT const char* myqq_get_group_name( int qid, uint int_uid )
{
	myqq_group_t *g = group_get2( qid, int_uid, 0 );
	if( g == NULL )
		return "None";
	else
		return g->name;
}

//interface for getting group information
/* Output style:
*/
EXPORT int myqq_get_group_info( int qid, uint int_uid, char* buf, int size )
{
	char cate_str[4][10] = {"Classmate", "Friend", "Workmate", "Other" };
	myqq_group_t *g = group_get2( qid, int_uid, 0 );
	int len;
	if( size < 256 )
		return -EFAULT;
	if( g == NULL )
		return -EFAULT;
	len = sprintf( buf, 	"BuddyName           %-32s\n"
				"InternalID          %-32d\n"
				"ExternalID          %-32d\n"
				"Type                %-32s\n"
				"AuthenMethod        %-32s\n"
				"Category            %-32s\n"
				"Owner               %-32d\n"
				"Members             %-32d\n"
				"Description         \n%-32s\n"
				"Notice              \n%-32s\n",
		g->name, g->int_uid, g->ext_uid, (g->type==0x01)?"Normal":"Special",
		(g->auth_type==0x01)?"No": (g->auth_type==0x02)?"Yes":
			(g->auth_type==0x03)?"RejectAll":"Unknown",
		g->category > 3 ? cate_str[3] : cate_str[(int)g->category],
		g->mgr_uid, g->member_count, g->desc, g->notice );
	return len;
}

