/*
 *  packet.c
 *
 *  A small packet manager
 *
 *  Copyright (C) 2008  Huang Guan
 *
 *  2008-01-31 Created. Note! there are some bugs in this file. To avoid race
 *  condition, we should use semaphore in every packet item.
 *
 *
 *  Description: This file mainly includes the functions about 
 *               Allocating a Packet
 *               Sending and Receiving Packet
 *               Searching and Matching Packets
 */


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#ifdef __WIN32__
#include <winsock.h>
#include <wininet.h>
#else
#include <sys/socket.h>
#include <arpa/inet.h>
#endif
#include "myqq.h"
#include "debug.h"
#include "mysocket.h"
#include "packet.h"
#include "protocol.h"
#include "rlist.h"



void packet_list_init( int qid )
{
	memset( &MYQQ(qid).plist, 0, sizeof( MYQQ(qid).plist ) );
	pthread_mutex_init( &MYQQ(qid).plist.mutex, NULL );
	rlist_init( &MYQQ(qid).rlist, PACKET_BUF_SIZE );
}

void packet_list_cleanup( int qid )
{
	rlist_cleanup( &MYQQ(qid).rlist );
}

myqq_prot_packet_t* packet_alloc( int qid, int len, int type )
{
	int i;
	myqq_prot_packet_t* packet = (myqq_prot_packet_t* )malloc(len);
	if( packet==NULL )
	{
		DBG("packet==NULL");
		exit(1);
	}
	memset( packet, 0, len );
	if( type == PACKET_OUT )
	{
		packet->headflag = 0x02;
		packet->version = MYQQ(qid).version;
#ifdef __WIN32__	//ǳֵ⣡
		char ch[4], tmp;
		memcpy( ch, &MYQQ(qid).uid, 4 );
		tmp = ch[0];ch[0]=ch[3];ch[3]=tmp;
		tmp = ch[1];ch[1]=ch[2];ch[2]=tmp;
		memcpy( &packet->uid, ch, 4 ) ;
#else
		packet->uid = htonl( MYQQ(qid).uid );
#endif
		packet->seqno = 0; 
	}
	//add to list
	pthread_mutex_lock( &MYQQ(qid).plist.mutex );
//---
	i = MYQQ(qid).plist.count;
	MYQQ(qid).plist.items[i].packet = packet;
	MYQQ(qid).plist.items[i].type = type;
	MYQQ(qid).plist.items[i].send_times = 0;
	MYQQ(qid).plist.items[i].time = 0;
	MYQQ(qid).plist.items[i].packet_len = len;
	MYQQ(qid).plist.count ++;
//---
	pthread_mutex_unlock( &MYQQ(qid).plist.mutex );
	
	return packet;
}

//free one or two packets and remove it from packet list.
void packet_free( int qid, myqq_prot_packet_t* packet )
{
	//search
	int i;
	pthread_mutex_lock( &MYQQ(qid).plist.mutex );
	for( i=0; i < MYQQ(qid).plist.count; i++ )
		if( MYQQ(qid).plist.items[i].packet->seqno == packet->seqno )
		{
			//remove it
			if( i != MYQQ(qid).plist.count-1 )
				MYQQ(qid).plist.items[i] = MYQQ(qid).plist.items[MYQQ(qid).plist.count-1];
			MYQQ(qid).plist.count--;
			i=-1; //redo??
		}
	pthread_mutex_unlock( &MYQQ(qid).plist.mutex );
	free( packet );
}

//match two packets. one for sending one for receiving
myqq_prot_packet_t* packet_match( int qid, myqq_prot_packet_t* p1 )
{
	myqq_prot_packet_t* p2;
	int i;
	pthread_mutex_lock( &MYQQ(qid).plist.mutex );
	for( i=0; i < MYQQ(qid).plist.count; i++ ){
		p2 = MYQQ(qid).plist.items[i].packet;
		if( p1 == p2 ) continue;
		if( p1->seqno == p2->seqno && p1->command == p2->command ) //matched
		{
			pthread_mutex_unlock( &MYQQ(qid).plist.mutex );
			return p2;
		}
	}
	pthread_mutex_unlock( &MYQQ(qid).plist.mutex );
	return NULL;
}

myqq_packet_item_t* packet_get_item( int qid, myqq_prot_packet_t* p1 )
{
	myqq_prot_packet_t* p2;
	int i;
	for( i=0; i < MYQQ(qid).plist.count; i++ ){
		p2 = MYQQ(qid).plist.items[i].packet;
		if( p1 == p2 ){
			return &MYQQ(qid).plist.items[i];
		}
	}
	return NULL;
}

int packet_send( int qid, myqq_prot_packet_t* packet, int len )
{
	int ret;
	if( !MYQQ(qid).sock_im ) return -EFAULT;
	myqq_packet_item_t* item = packet_get_item( qid, packet );
	if( item == NULL )
	{
		DBG("Impossible."); exit(1);
	}
	item->time = time(NULL);
	item->packet_len = len;
	item->send_times ++;
	ret = mysock_send( MYQQ(qid).sock_im, (char*)packet, len );
	if( ret!=len )
		DBG("packet send failed.ret:%d", ret);
	return ret;
}

int packet_recv( int qid, myqq_prot_packet_t* packet, int len )
{
	int ret, seqcmd;
	if( !MYQQ(qid).sock_im ) return -EFAULT;
	ret = mysock_recv( MYQQ(qid).sock_im, (char*)packet, len );
	if( ret<=5 )
		DBG("packet recv failed.");
	seqcmd = (ntohs( packet->seqno )<<16) ;//| ntohs( packet->command );
	if( rlist_find( &MYQQ(qid).rlist, seqcmd) )
	{
		DBG("same cmd:%d seqno:%d", ntohs(packet->command), ntohs(packet->seqno) );
		return -EFAULT; //give it up
	}
	rlist_append( &MYQQ(qid).rlist, seqcmd );
	myqq_packet_item_t* item = packet_get_item( qid, packet );
	if( item == NULL )
	{
		DBG("IMPOSSIBLE.");
		return -EFAULT;
	}
	item->time = time(NULL);
	return ret;
}

//return a time out packet
myqq_prot_packet_t* packet_timeout( int qid, int secs, int* len, int* send_times )
{
	int i;
	time_t t;
	pthread_mutex_lock( &MYQQ(qid).plist.mutex );
	for( i=0; i < MYQQ(qid).plist.count; i++ ){
		t = MYQQ(qid).plist.items[i].time;
		if( t == 0 ) continue; //not sent packet.
		if( time(NULL)-t >= secs ) //matched
		{
			if( MYQQ(qid).plist.items[i].type == PACKET_IN )
			{
				DBG("Warning: packet not handled. cmd:0x%x", 
					ntohs(MYQQ(qid).plist.items[i].packet->command) );
			}else if( MYQQ(qid).plist.items[i].type == PACKET_OUT )
			{
				*len = MYQQ(qid).plist.items[i].packet_len;
				*send_times = MYQQ(qid).plist.items[i].send_times;
				pthread_mutex_unlock( &MYQQ(qid).plist.mutex ); 
				DBG("wait to send cmd: %x  seqno: %x", 
					ntohs( MYQQ(qid).plist.items[i].packet->command ),
					ntohs( MYQQ(qid).plist.items[i].packet->seqno ) );
				//!!: security problem: packet might be deleted while using by another
				// we should use semaphore.
				return MYQQ(qid).plist.items[i].packet;
			}
		}
	}
	pthread_mutex_unlock( &MYQQ(qid).plist.mutex );
	return NULL;
}

